"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-10
Last updated: 2023-05-27
"""

import logging
import pandas as pd

from datetime import datetime

from synd.utils import create_timestamp
from synd.datautil import (
    TabularSampler,
    TabularTransformer,
    DTypeTransformer,
)

from synd.typing import *
from synd.datasets.base import Dataset

log = logging.getLogger(__name__)

class SingleTable(Dataset):
    """ """

    def __init__(self,
        data: pd.DataFrame,
        metadata: Dict,
        *,
        log_frequency: Boolean = True,
        discrete_columns: Optional[List[String]] = None,
        max_bgm_clusters: Integer = 10,
        weight_threshold: Float = 0.005,
        name: Optional[String] = None,
        **kwargs: Dict,
    ):
        super(SingleTable, self).__init__(**kwargs)

        if discrete_columns is None:
            discrete_columns = []

        self._data = data
        self._metadata = metadata

        self._log_frequency = log_frequency
        self._discrete_columns = discrete_columns
        self._max_bgm_clusters = max_bgm_clusters
        self._weight_threshold = weight_threshold

    def __len__(self) -> Integer:
        return self._data.shape[0]

    def fit(self, transformer_kwargs={}, sampler_kwargs={}):
        """ """

        dtype_transformer = DTypeTransformer()
        log.debug(f'fitting `{dtype_transformer}` with real data...')
        dtype_transformer.fit(
            self._data,
            self._metadata,
            discrete_columns=self._discrete_columns,
        )

        transformer = TabularTransformer(
            max_bgm_clusters=self._max_bgm_clusters,
            weight_threshold=self._weight_threshold,
            **transformer_kwargs,
        )

        log.debug(f'created `{transformer}`')
        log.debug(f'fitting `{transformer}` with real data...')
        transformer.fit(
            self._data,
            discrete_columns=self._discrete_columns,
        )

        log.debug(f'transforming real data with `{transformer}`...')
        transformed_data = transformer.transform(self._data)

        sampler = TabularSampler(
            transformed_data,
            transformer._col_act_info,
            log_frequency=self._log_frequency,
            **sampler_kwargs,
        )

        log.debug(f'created `{sampler}` with transformed data')
        self._transformer = transformer
        self._dtype_transformer = dtype_transformer
        self._sampler = sampler
        self._fitted = True

    def is_fitted(self) -> Boolean:
        return hasattr(self, '_fitted')
    
    @property
    def name(self) -> String:
        return self._name
    
    @property
    def data(self) -> pd.DataFrame:
        return self._data
    
    @property
    def metadata(self) -> Dict:
        return self._metadata
    
    @property
    def transformer(self) -> Optional[TabularTransformer]:
        return getattr(self, '_transformer', None)

    @property
    def sampler(self) -> Optional[TabularSampler]:
        return getattr(self, '_sampler', None)

