"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-27
Last updated: 2023-05-28
"""

import numpy as np
import pandas as pd

from synd.info import ColumnDTypeInfo
from synd.datautil.transformer import Transformer
from synd.typing import *

class DTypeTransformer(Transformer):
    """ """

    def fit(self,
        data: pd.DataFrame,
        metadata: Dict,
        discrete_columns: Tuple[String, ...],
        **kwargs: Dict,
    ):
        """ """

        col_dtype_info = []
        for col in data.columns:
            col_info = self._fit_fn[
                'discrete' if col in discrete_columns else 'continuous'
            ](data[col], metadata['columns'][col], col)
            col_dtype_info.append(col_info)

        self._fitted = True
        self._col_dtype_info = col_dtype_info

    def _fit_discrete(self,
        data: pd.Series,
        metadata: Dict,
        column: String,
        **kwargs: Dict,
    ):
        """ """

        dtype = metadata.get('sdtype', None)
        regex = None

        if dtype == 'id':
            regex = metadata.get('regex_format')

        return ColumnDTypeInfo(
            _name=column,
            _type='discrete',
            _dtype=dtype,
            _regex=regex,
            _n_decimals=None,
        )

    def _fit_continuous(self,
        data: pd.Series,
        metadata: Dict,
        column: String,
        **kwargs: Dict,
    ):
        """ """

        dtype = data.dtype
        n_decimals = 0
        if dtype not in ['int16', 'int32', 'int64']:
            n_decimals = max([len(str(d).split('.')[-1]) for d in data])

        return ColumnDTypeInfo(
            _name=column,
            _type='continuous',
            _dtype=dtype,
            _regex=None,
            _n_decimals=n_decimals,
        )

    def transform(self, data: pd.DataFrame, **kwargs: Dict) -> pd.DataFrame:
        """ """

        self.assert_is_fitted()

        col_names = []
        data_list = []
        for col_info in self._col_dtype_info:
            col_name = col_info._name
            col_names.append(col_name)
            data_list.append(
                self._transform_fn[col_info._type](
                    data[col_name], col_info,
                ))

        data = pd.DataFrame({
            col: data_list[i] for i, col in enumerate(col_names)
        })

        return data

    def _transform_discrete(self,
        data: pd.Series,
        info: ColumnDTypeInfo,
        **kwargs: Dict,
    ):
        """ TODO: implement id parsing and text validation? """
        return data

    def _transform_continuous(self,
        data: pd.Series,
        info: ColumnDTypeInfo,
        **kwargs: Dict,
    ):
        """ """

        n_decimals = info._n_decimals
        if n_decimals == 0:
            return data.astype(np.int32)

        return data.round(decimals=info._n_decimals)

