"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
coalphaes of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
coalphaes or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-13
Last updated: 2023-05-28
"""

from __future__ import annotations

import logging
import pandas as pd
import numpy as np

from synd.datautil.sampler import Sampler
from synd.datautil.info import ActivationFnInfo
from synd.typing import *
from synd.utils import is_discrete_column

log = logging.getLogger(__name__)

class TabularSampler(Sampler):
    """ Sampler derivative for sampling tabular data of mixed types, extended to sample
    the conditional vector used with the CTGAN model [1].

    The conditional vector `c` is created in the following manner:
        0) For each item in the batch:
            1) Uniformly at random, sample a discrete column `D_i*`.
            2) Sample a cateogory `k*` to condition on randomly according to the log
               frequency of the categories in the column `D_i*`.
            3) Create a one-hot encoded masking vector `m` where the k*'th item is marked
               as as `1` and all others `0`.
        4) Return a batch matrix of size `(batch_size, max_cat_len)` where `max_cat_len`
           is the largest number of categories `k` that any discrete column `D_i` has.

    For example, given the three discrete columns:
    D_1 = {cat, dog, capybara}, D_2 = {29, 49}, D_3 = {green, blue, yellow, red}
    we could sample `D_i* = 1`, and then from that `k* = capybara`, meaning, our
    conditional vector `c` is subject to `(D_1 = capybara)`, which would yield the
    following conditional vector: `c = [0, 0, 1, 0, 0, 0, 0, 0, 0]`.

    "During training the conditional generator is free to produce any set of one-hot
    discrete vectors `{d_1, ..., d_n}` which is not specifically subject to `(D_i* = k*)`.
    This is implemented by uniformly at random sampling entropy which is compared against
    the cumulative sum of the log probabilities of `D_i*` [1].

    Parameters
    ----------
    data: pd.DataFrame
        The data to sample from. Expects it to have been transformed by `Transformer`
        to be numerical modality only. For specific ways of transforming mixed types data,
        see implementation of e.g. `TabularTransformer`.
    info: list
        A list where each item is a list of the activation function information needed to
        forward- and backpropagate the data through a neural network. The
        `ActivationFnInfo` dataclass contains information about the number of dimensions
        and what activation function should be applied to those dimensions.
    log_frequency: bool
        Whether or not to sample the data and conditional vector according to the
        log-frequency of each discrete category. This helps with data imbalance.
        During inference, this is disabled. Defaults to `True`.

    [1] Lei Xu et al., 2019

    """

    def __init__(self,
        data: pd.DataFrame,
        info: List[List[ActivationFnInfo]],
        *,
        log_frequency: Boolean = True,
        **kwargs: Dict,
    ):
        super(TabularSampler, self).__init__()

        # row id for each categorical value in each discrete column
        # e.g., rid_by_cat_cols[a][b] is a list of all rows with 
        # category b in discrete column a
        rid_by_cat_cols = []
        st = 0

        for col_info in info:
            if not is_discrete_column(col_info):
                st += sum([act_info._dim for act_info in col_info])
                continue

            act_info = col_info[0]
            ed = st + act_info._dim

            # get the indices of the elements that are non-zero
            rid_by_cat = [
                np.nonzero(data[:, st + j])[0]
                for j in range(act_info._dim)
            ]

            rid_by_cat_cols.append(rid_by_cat)
            st = ed

        assert st == data.shape[1]

        max_category = max([
            col_info[0]._dim
            for col_info in info
            if is_discrete_column(col_info)
        ], default=0)

        n_categories = sum([
            col_info[0]._dim for col_info
            in info if is_discrete_column(col_info)
        ])

        n_discrete_columns = sum([
            1 for col_info in info
            if is_discrete_column(col_info)
        ])

        d_col_cond_st = np.zeros(n_discrete_columns).astype(np.int32)
        d_col_n_cat = np.zeros(n_discrete_columns).astype(np.int32)
        d_col_cat_prob = np.zeros((n_discrete_columns, max_category))
        d_col_cat_prob_log = np.zeros((n_discrete_columns, max_category))

        st = 0
        curr_id = 0
        curr_cond_st = 0
        for col_info in info:
            if not is_discrete_column(col_info):
                st += sum([act_info._dim for act_info in col_info])
                continue
            
            act_info = col_info[0]
            ed = st + act_info._dim

            cat_freq = np.sum(data[:, st:ed], axis=0)
            cat_freq_log = np.log(cat_freq + 1)

            cat_prob = cat_freq / np.sum(cat_freq)
            cat_prob_log = cat_freq_log / np.sum(cat_freq_log)

            d_col_cat_prob[curr_id, :act_info._dim] = cat_prob
            d_col_cat_prob_log[curr_id, :act_info._dim] = cat_prob_log
            d_col_cond_st[curr_id] = curr_cond_st
            d_col_n_cat[curr_id] = act_info._dim

            st = ed
            curr_id += 1
            curr_cond_st += act_info._dim

        self._data = data
        self._log_frequency = log_frequency
        self._rid_by_cat_cols = rid_by_cat_cols

        self._n_rows = len(data)
        self._n_categories = n_categories
        self._n_discrete_columns = n_discrete_columns

        self._d_col_cond_st = d_col_cond_st
        self._d_col_n_cat = d_col_n_cat
        self._d_col_cat_prob = d_col_cat_prob
        self._d_col_cat_prob_log = d_col_cat_prob_log

    def __len__(self) -> Integer:
        return len(self._data)
        
    @property
    def n_categories(self) -> Integer:
        return self._n_categories

    @property
    def n_discrete_columns(self) -> Integer:
        return self._n_discrete_columns
    
    def sample_condvec(self,
        batch_size: Integer,
    ) -> Tuple[Optional[np.ndarray], ...]:
        """ """
        if self._n_discrete_columns == 0:
            return (None, ) * 4

        i = np.random.choice(
            np.arange(self._n_discrete_columns),
            batch_size,
        ).astype(np.int32)

        m = np.zeros((batch_size, self._n_discrete_columns)).astype(np.int32)
        m[np.arange(batch_size), i] = 1

        probs = self._d_col_cat_prob_log[i].cumsum(axis=1)

        if not self._log_frequency:
            probs = self._d_col_cat_prob[i].cumsum(axis=1)
        
        entropy = np.expand_dims(
            np.random.rand(batch_size),
            axis=1,
        )

        k = (probs > entropy).argmax(axis=1)
        st = self._d_col_cond_st[i] + k

        c = np.zeros((batch_size, self._n_categories)).astype(np.int32)
        c[np.arange(batch_size), st] = 1

        return (c, m, i, k)
    
    def sample_original_condvec(self,
        batch_size: Integer,
    ) -> Optional[np.ndarray]:
        """ Sample the conditional vector `c` from the original column category
        frequency probabilities. This method is used during inference, when sampling
        new synthetic data, after training. Use `.sample_condvec()` during training.

        Parameters
        ----------
        batch_size: int
            The number of conditional vectors `c` to create.

        Returns
        -------
        c: np.ndarray | None
            A matrix containing all of the conditional vectors. The matrix will have
            shape `(batch_size, self._n_categories)`. Or it can return `None`, only
            happens if the data does not contain any discrete columns.

        """

        if self._n_discrete_columns == 0:
            return None

        i = np.random.choice(
            np.arange(self._n_discrete_columns),
            batch_size,
        ).astype(np.int32)

        probs = self._d_col_cat_prob[i].cumsum(axis=1)
        
        entropy = np.expand_dims(
            np.random.rand(batch_size),
            axis=1,
        )

        k = (probs > entropy).argmax(axis=1)
        k = self._d_col_cond_st[i] + k

        c = np.zeros((batch_size, self._n_categories)).astype(np.int32)
        c[np.arange(batch_size), k] = 1 

        return c
    
    def sample_data(self,
        n_rows: Integer,
        *,
        cols: Optional[np.ndarray] = None,
        cats: Optional[np.ndarray] = None,
        **kwargs: Dict,
    ) -> np.ndarray:
        """ """

        if cols is None or cats is None:
            return self._data[
                np.random.randint(len(self), size=(n_rows, )),
            ]

        sampled = [
            np.random.choice(self._rid_by_cat_cols[i][k])
            for i, k in zip(cols, cats)
        ]

        return self._data[sampled]
    
