"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
coalphaes of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
coalphaes or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-13
Last updated: 2023-05-21
"""

from __future__ import annotations

import warnings
import logging
import pandas as pd
import numpy as np
import torch

from sklearn.preprocessing import OneHotEncoder
from sklearn.mixture import BayesianGaussianMixture

from synd.datautil.transformer import Transformer
from synd.info import (
    ColumnTransformInfo,
    ActivationFnInfo,
)

from synd.typing import *
from synd.utils import get_rng_state

log = logging.getLogger(__name__)

class TabularTransformer(Transformer):
    """ """

    def __init__(self, *,
        max_bgm_clusters: Integer = 10,
        weight_threshold: Float = 0.005,
        **kwargs: Dict,
    ):
        super(TabularTransformer, self).__init__(cls_=self, **kwargs)

        self._max_bgm_clusters = max_bgm_clusters
        self._weight_threshold = weight_threshold

    def fit(self,
        data: pd.DataFrame,
        discrete_columns: Tuple[String, ...],
        **kwargs: Dict,
    ):
        """ """
        n_output_dims = 0
        col_act_info = []
        col_transform_info = []

        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            for col in data.columns:
                col_info = self._fit_fn[
                    'discrete' if col in discrete_columns else 'continuous'
                ](data[col], col)
                n_output_dims += col_info._dims
                col_act_info.append(col_info._act)
                col_transform_info.append(col_info)

        self._fitted = True
        self._n_output_dims = n_output_dims
        self._col_act_info = col_act_info
        self._col_transform_info = col_transform_info

    def _fit_discrete(self,
        data: pd.Series,
        column: String,
        *,
        sparse_output: Boolean = False,
        handle_unknown: String = 'error',
        **kwargs: Dict,
    ) -> ColumnTransformInfo:
        """ """

        data = data.to_numpy().reshape(-1, 1)
        ohe = OneHotEncoder(
            sparse_output=sparse_output,
            handle_unknown=handle_unknown,
        )

        ohe.fit(data)
        n_categories = ohe.transform(data).shape[1]

        return ColumnTransformInfo(
            _name=column,
            _type='discrete',
            _dims=n_categories,
            _act=[ActivationFnInfo(n_categories, 'softmax')],
            _fit=ohe,
        )

    def _fit_continuous(self,
        data: pd.Series,
        column: String,
        *,
        n_init: Integer = 1,
        init_params: String = 'random_from_data',
        weight_concentration_prior_type: String = 'dirichlet_process',
        weight_concentration_prior: Float = 1e-3,
        **kwargs: Dict,
    ) -> ColumnTransformInfo:
        """ """

        bgm = BayesianGaussianMixture(
            n_components=self._max_bgm_clusters,
            n_init=n_init,
            init_params=init_params,
            weight_concentration_prior_type=weight_concentration_prior_type,
            weight_concentration_prior=weight_concentration_prior,
            random_state=get_rng_state()['numpy'][1][0],
        )

        bgm.fit(data.to_numpy().reshape(-1, 1))

        valid_components = bgm.weights_ > self._weight_threshold
        bgm._valid_components = valid_components
        n_components = sum(valid_components)

        return ColumnTransformInfo(
            _name=column,
            _type='continuous',
            _dims=1 + n_components,
            _act=[ActivationFnInfo(1, 'tanh'), ActivationFnInfo(n_components, 'softmax')],
            _fit=bgm,
        )

    def transform(self, data: pd.DataFrame) -> np.ndarray:
        """ """

        self.assert_is_fitted()

        data_list = []
        for transform_info in self._col_transform_info:
            column = transform_info._name
            data_list.append(self._transform_fn[transform_info._type](
                    data[column], transform_info,
            ))

        return np.concatenate(data_list, axis=1)
    
    def _transform_discrete(self,
        data: pd.Series,
        info: ColumnTransformInfo,
        **kwargs: Dict,
    ) -> np.ndarray:
        """ """
        return info._fit.transform(data.to_numpy().reshape(-1, 1)).astype(np.float32)
    
    def _transform_continuous(self,
        data: pd.Series,
        info: ColumnTransformInfo,
        *,
        eps: Float = 1e-6,
        **kwargs: Dict,
    ) -> np.ndarray:
        """ 
        Mode-specific normalization:
        https://arxiv.org/abs/1907.00503
        """

        data = data.to_numpy().reshape(-1, 1)
        bgm = info._fit
        means = bgm.means_.reshape(1, -1)
        stds = np.sqrt(bgm.covariances_).reshape(1, -1)
        
        normalized_data = ((data - means) / (4 * stds))[:, bgm._valid_components]
        probs = bgm.predict_proba(data)[:, bgm._valid_components]

        selected = np.array([
            np.random.choice(
                np.arange(bgm._valid_components.sum()),
                p=((probs[i] + eps) / (probs[i] + eps).sum()),
            ) for i in range(data.shape[0])
        ]).astype(np.int32)

        normalized = normalized_data[np.arange(data.shape[0]), selected].reshape(-1, 1)
        normalized = np.clip(normalized, -0.99, 0.99)[:, 0]

        transformed = np.stack([normalized, selected], axis=1)

        output = np.zeros((transformed.shape[0], info._dims))
        output[:, 0] = transformed[:, 0]

        index = transformed[:, 1].astype(np.int32)
        output[np.arange(index.shape[0]), index + 1] = 1.0

        return output

    def inverse_transform(self,
        data: np.ndarray,
        **kwargs: Dict,
    ) -> pd.DataFrame:
        """ """

        if not self.is_fitted():
            raise ValueError(
                f'`{self.__class__.__name__}` has not been fitted, please run `.fit()` '
                f'with your tabular dataset before attempting to inverse_transform.'
            )

        st = 0
        data_list = []
        col_names = []
        for transform_info in self._col_transform_info:
            dims = transform_info._dims
            ed = st + dims
            col_data = data[:, st:ed]

            data_list.append(self._inverse_transform_fn[transform_info._type](
                col_data,
                transform_info,
            ))

            col_names.append(transform_info._name)
            st = ed

        data = np.hstack(data_list)
        data = pd.DataFrame({
            col: data_list[i].flatten() for i, col in enumerate(col_names)
        })

        return data
    
    def _inverse_transform_discrete(self,
        data: np.ndarray,
        info: ColumnTransformInfo,
        **kwargs: Dict,
    ) -> np.ndarray:
        """ """
        return info._fit.inverse_transform(data)

    def _inverse_transform_continuous(self,
        data: np.ndarray,
        info: ColumnTransformInfo,
        **kwargs: Dict,
    ) -> np.ndarray:
        """ """
        
        selected = np.argmax(data[:, 1:], axis=1)
        normalized = np.clip(data[:, 0], -1, 1).reshape(-1, 1)

        bgm = info._fit
        valid_components = bgm._valid_components

        means = bgm.means_.reshape(-1, 1)[valid_components][selected]
        stds = np.sqrt(bgm.covariances_).reshape(-1, 1)[valid_components][selected]

        inverse_data = normalized * 4 * stds + means
        return inverse_data

