"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-14
Last updated: 2023-05-28
"""

import contextlib
from functools import wraps

import numpy as np
import torch

from synd.base import Base
from synd.typing import *

@contextlib.contextmanager
def set_rng_states(
    state: Tuple[np.random.RandomState, torch.Generator],
    fn: Callable,
):
    """ """

    original_np_state = np.random.get_state()
    original_torch_state = torch.get_rng_state()

    np_rng, torch_rng, = state

    try:
        yield
    finally:

        curr_np_rng = np.random.RandomState()
        curr_np_rng.set_state(np_rng.get_state())
        curr_torch_rng = torch.Generator()
        curr_torch_rng.set_state(torch_rng.get_state())

        fn(state=(
            curr_np_rng,
            curr_torch_rng,
        ))

        np.random.set_state(original_np_state)
        torch.set_rng_state(original_torch_state)

def random_state(fn: Callable) -> Callable:
    """ """

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        if self._rng_state is None:
            return fn(self, *args, **kwargs)

        else:
            with set_rng_states(self._rng_state, self._set_rng_state):
                return fn(self, *args, **kwargs)

    return wrapper

class RandomState(Base):
    """ """

    def __init__(self, *args, **kwargs):
        super(RandomState, self).__init__(*args, **kwargs)
        self._set_rng_state(**kwargs)
    
    def _set_rng_state(self, *,
        state: Union[Integer, Tuple[np.random.RandomState, torch.Generator]] = 1337,
        **kwargs: Dict,
    ):
        """ Set the state for all relevant random number generators. """

        if isinstance(state, int):
            self._rng_state = (
                np.random.RandomState(seed=state),
                torch.Generator().manual_seed(state),
            )
        elif (
            isinstance(state, tuple) and
            isinstance(state[0], np.random.RandomState) and
            isinstance(state[1], torch.Generator)
        ):
            self._rng_state = state
        else:
            raise TypeError(
                f'The provided state should either be `int` or (`np.random.RandomState`, '
                f'`torch.Generator`). Function got: {state}.'
            )
        
