"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
coalphaes of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
coalphaes or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-08
Last updated: 2023-05-21
"""

from __future__ import annotations

import torch
import torch.nn as nn

from synd.typing import *

class Critic(nn.Module):
    """
    Critic neural network for Wasserstein PacGAN.

    [1] Martin Arjovsky et al., 2017 `https://arxiv.org/abs/1701.07875`
    [2] Zinan Lin et al., 2017 `https://arxiv.org/abs/1712.04086`
    [3] Ishaan Gulrajani., 2017 `https://arxiv.org/abs/1704.00028`
    """

    def __init__(self,
        input_dim: Integer,
        critic_dims: Tuple[Integer, ...],
        *,
        negative_slope: Float = 0.2,
        dropout: Float = 0.5,
        pac: Integer = 10,
        device: Union[torch.device, String] = 'cpu',
        **kwargs: Dict,
    ):
        super(Critic, self).__init__()

        self._input_dim = input_dim
        self._critic_dims = critic_dims
        self._negative_slope = negative_slope 
        self._dropout = dropout
        self._pac = pac
        self._device = device

        dims = [input_dim * pac] + list(critic_dims)

        sequence = []
        for in_, out in zip(dims[:-1], dims[1:]):
            sequence += [nn.Linear(in_, out)]
            sequence += [nn.LeakyReLU(negative_slope)]
            sequence += [nn.Dropout(dropout)]

        sequence += [nn.Linear(dims[-1], 1)]
        encoder = nn.Sequential(*sequence).to(device)
        self._encoder = encoder 

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert x.size()[0] % self._pac == 0, \
        f'Batch size of input tensor must be divisible by the pac size. ' \
        f'Your shape is {x.shape}, and {x.size()[0]} is not divisible by {self._pac}.'

        return self._encoder(x.view(-1, self._input_dim * self._pac))

    @classmethod
    def no_pac(cls, *args, **kwargs) -> Critic:
        return cls(*args, pac=1, **kwargs)

