"""
MIT License

Copyright (c) 2023 Wilhelm Ågren

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

File created: 2023-05-09
Last updated: 2023-05-21
"""

from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F

from synd.typing import *

class Residual(nn.Module):
    """ Residual block for a neural network [1]. 
    
    [1] Kaiming He et al., 2015

    """

    def __init__(self,
        in_: Integer,
        out: Integer,
        *,
        skip: Boolean = False,
        device: Union[String, torch.tensor] = 'cpu',
        **kwargs: Dict,
    ):
        super(Residual, self).__init__()

        self._in_dim = in_
        self._out_dim = out 
        self._skip = skip
        self._device = device

        sequence = []
        sequence += [
            nn.Linear(in_, in_, bias=False),
            nn.BatchNorm1d(in_),
            nn.ReLU(),
        ]
        sequence += [
            nn.Linear(in_, out, bias=False),
            nn.BatchNorm1d(out),
        ]

        block = nn.Sequential(*sequence).to(device)
        self._block = block
        self._skip = nn.Linear(in_, out) if not skip else nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """ """
        h = self._block(x)
        return F.relu(h + self._skip(x))

